﻿// Webcam Signature 1.3
// Created by Eric Wong
// Copyright 2008
// This is open source software licensed under the GNU GENERAL PUBLIC LICENSE V3.
// You must share your source code if you utilize this code in your program.
// I wish to promote the development of computer vision so that it is easy to use
// for casual users and programmers, as well as advanced ones. Right now, many of the
// techniques used in computer vision are quite archaic and without easy to understand
// coding. Please share.

// Please link back to the website below if you use this code.
// Feel free to contact me at VideoInPicture@gmail.com if there are any questions
// or if you would like to help develop this project.
// June 20th, 3008
// http://webcamsignature.wikidot.org
//
// This program allows a user to write his/her name with a webcam and a colored object.
// The program tracks colors to determine what to draw and creates an array of points (ArrayList SignaturePoints)
// based on the path that the colored object moves.
// 
// The program creates smooth signatures by drawing cardinal splines through the array of points
// where possible.
//
// The original purpose of Webcam Signature is to create a mouse moving program
// that just utilizes a person's hands with no other objects needed . This is still 
// the long term goal of the project.

// The tracking code is located in the region "Core Code".
// The most important function is private int GetPoint(....)
// This function applies the background filtering and looks at the pixels in the images
// captured from the webcam to find the pixels with the wanted colors.
// It then averages the location of the valid pixels to determine at what point on the screen
// the user is referencing.

// The program utilizes DShowNet to interface with the webcam. It uses a really
// old version (1.0), because I found it to be faster than the new version (2.0).
// However, I have not extensively tested this and and likely incorporate version 2.0
// into the program later.

// The program is hard coded for a 640x480 webcam resolution at 20fps.
// User defined resolutions will be introduced in version 2.0.

// This version of Webcam Signature introduces background filtering.

// In the next version, we need to improve the tracking of the program near the edges of the captured
// webcam images. We also need to code the dialation and erosion functions to reduce noise.
// Right now, the program tracks really accurately when there is even lighting in the room and the color
// of the object stands out. It also works really well when you use a flashlight in place of a pen.

using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.Data;
using System.Runtime.InteropServices;
using System.Threading;
using System.IO;
using SampleGrabberNET;
using DShowNET;
using DShowNET.Device;
using FancyColorDialog;


namespace WebCam_Signature
{
    public partial class MainForm : Form, ISampleGrabberCB
    {
        #region Constants & Variables

        // The following three values are hard coded because the program is not set up to 
        // handle other resolutions and it simplified coding for a 1.0 release.
        // User defined resolutions will be coded in Webcam Signature 2.0
        const int iHeight = 480;
        const int iWidth = 640;
        const int FrameRate = 20;
    
        int RedL, RedH, GreenL, GreenH, BlueL, BlueH;
        float PenThickness;
        bool isKeyDown;
        bool DoneProcessing;
        public bool FormActivated;
        int DrawingWidth, DrawingHeight;
        bool WebcamMouseMode;

        // Variables used to store the Bitmap images of the drawing area
        Bitmap SignaturePixels;
        Bitmap ProcessPixels;
        UnsafeBitmap WebCamBitmap;

        // Variables to store the color for the signature and the signature background color
        PixelData SignatureColor;
        Color SignatureColorSystem;
        PixelData backColor;
        Color backColorSystem;

        // SignaturePoints stores the points recorded for the movement path of the colored object.
        // Undo and Redo are arrays of SignaturePoints at various states.
        ArrayList SignaturePoints = new ArrayList();
        ArrayList Undo = new ArrayList();
        ArrayList Redo = new ArrayList();

        public FineTuneTrackColorsForm FineTuneTC;
        ColorButton PenColorButton = new ColorButton();
        ColorButton BackgroundColorButton = new ColorButton();

        // Variables used to store data for the background average colors
        // and the behavior for background filtering
        double[, ,] BackgroundAverage;
        int RequiredNumBackgroundFrames;
        int NumBackgroundFramesCaptured;
        int BackgroundDiffThreshold;
        int BackgroundCalibrateTime;
        int NumBackgroundFramesRequiredCalibration;
        bool NoBackgroundFilter;
        bool UseBackgroundStaticFilter;
        bool UseBackgroundDynamicFilter;

        #endregion

        #region Mouse Control

        private const UInt32 MOUSEEVENTF_LEFTDOWN = 0x0002;
        private const UInt32 MOUSEEVENTF_LEFTUP = 0x0004;
        private const UInt32 MOUSEEVENTF_RIGHTDOWN = 0x0008;
        private const UInt32 MOUSEEVENTF_RIGHTUP = 0x0010;
        private const UInt32 MOUSEEVENTF_MIDDLEDOWN = 0x20;
        private const UInt32 MOUSEEVENTF_MIDDLEUP = 0x40;
        private const UInt32 MOUSEEVENTF_WHEEL = 0x800;

        [DllImport("user32.dll")]
        private static extern void mouse_event(UInt32 dwFlags, UInt32 dx, UInt32 dy, UInt32 dwData, IntPtr dwExtraInfo);

        public static void SendClick(Point location)
        {
            Cursor.Position = location;
            mouse_event(MOUSEEVENTF_LEFTDOWN, 0, 0, 0, new IntPtr());
            mouse_event(MOUSEEVENTF_LEFTUP, 0, 0, 0, new IntPtr());
        }

        #endregion

        #region WebCam API
        [DllImport("user32.dll")]
        static extern IntPtr GetForegroundWindow();

        // The following code in region WebCam API is a modified version of the image capturing code
        // from the first version of the DShowNet library. It utilizes DirectX 8.1 or higher.
        // DShowNet is an excellent wrapper to allow easy interfacing with the webcam.
        // Most other image capturing code utilizes the clipboard as the interface to the webcam.

        /// <summary> flag to detect first Form appearance </summary>
        private bool firstActive;

        /// <summary> base filter of the actually used video devices. </summary>
        private IBaseFilter capFilter;

        /// <summary> graph builder interface. </summary>
        private IGraphBuilder graphBuilder;

        /// <summary> capture graph builder interface. </summary>
        private ICaptureGraphBuilder2 capGraph;
        private ISampleGrabber sampGrabber;

        /// <summary> control interface. </summary>
        private IMediaControl mediaCtrl;

        /// <summary> event interface. </summary>
        private IMediaEventEx mediaEvt;

        /// <summary> video window interface. </summary>
        private IVideoWindow videoWin;

        /// <summary> grabber filter interface. </summary>
        private IBaseFilter baseGrabFlt;

        /// <summary> structure describing the bitmap to grab. </summary>
        private VideoInfoHeader videoInfoHeader;
        private bool captured = true;
        private int bufferedSize;

        /// <summary> buffer for bitmap data. </summary>
        private byte[] savedArray;

        /// <summary> list of installed video devices. </summary>
        private ArrayList capDevices;

        private const int WM_GRAPHNOTIFY = 0x00008001;	// message from graph

        private const int WS_CHILD = 0x40000000;	// attributes for video window
        private const int WS_CLIPCHILDREN = 0x02000000;
        private const int WS_CLIPSIBLINGS = 0x04000000;

        /// <summary> event when callback has finished (ISampleGrabberCB.BufferCB). </summary>
        private delegate void CaptureDone();

        private void InitializeCamera()
        {
            if (!DsUtils.IsCorrectDirectXVersion())
            {
                MessageBox.Show(this, "DirectX 8.1 or higher NOT installed!\n" + "Webcam functionality not available.\n", "Webcam Signature", MessageBoxButtons.OK, MessageBoxIcon.Stop);
                //this.Close(); 
                return;
            }

            if (!DsDev.GetDevicesOfCat(FilterCategory.VideoInputDevice, out capDevices))
            {
                MessageBox.Show(this, "No video capture devices found!\n" + "Please enable your webcam and restart the program.\n", "Webcam Signature", MessageBoxButtons.OK, MessageBoxIcon.Stop);
                //this.Close(); 
                return;
            }

            DsDevice dev = null;
            if (capDevices.Count == 1)
                dev = capDevices[0] as DsDevice;
            else
            {
                DeviceSelector selector = new DeviceSelector(capDevices);
                selector.ShowDialog(this);
                dev = selector.SelectedDevice;
            }

            if (dev == null)
            {
                this.Close(); return;
            }

            if (!StartupVideo(dev.Mon))
                this.Close();
        }

        /// <summary> start all the interfaces, graphs and preview window. </summary>
        bool StartupVideo(UCOMIMoniker mon)
        {
            int hr;
            try
            {
                if (!CreateCaptureDevice(mon))
                    return false;

                if (!GetInterfaces())
                    return false;

                if (!SetupGraph())
                    return false;

                if (!SetupVideoWindow())
                    return false;

                hr = mediaCtrl.Run();
                if (hr < 0)
                    Marshal.ThrowExceptionForHR(hr);

                // Don't care about tuners, at least for now.
                // This was annoying to have it pop up every time the program ran.
                //bool hasTuner = DsUtils.ShowTunerPinDialog(capGraph, capFilter, this.Handle);

                return true;
            }
            catch (Exception ee)
            {
                MessageBox.Show(this, "Could not start video stream\r\n" + ee.Message, "Webcam Signature", MessageBoxButtons.OK, MessageBoxIcon.Stop);
                return false;
            }
        }

        /// <summary> make the video preview window to show in videoPanel. </summary>
        bool SetupVideoWindow()
        {
            int hr;
            try
            {
                // Set the video window to be a child of the main window
                hr = videoWin.put_Owner(CameraPreviewPanel.Handle);
                if (hr < 0)
                    Marshal.ThrowExceptionForHR(hr);

                // Set video window style
                hr = videoWin.put_WindowStyle(WS_CHILD | WS_CLIPCHILDREN);
                if (hr < 0)
                    Marshal.ThrowExceptionForHR(hr);

                // Use helper function to position video window in client rect of owner window
                ResizeVideoWindow();

                // Make the video window visible, now that it is properly positioned
                hr = videoWin.put_Visible(DsHlp.OATRUE);
                if (hr < 0)
                    Marshal.ThrowExceptionForHR(hr);

                hr = mediaEvt.SetNotifyWindow(this.Handle, WM_GRAPHNOTIFY, IntPtr.Zero);
                if (hr < 0)
                    Marshal.ThrowExceptionForHR(hr);
                return true;
            }
            catch (Exception ee)
            {
                MessageBox.Show(this, "Could not setup video window\r\n" + ee.Message, "Webcam Signature", MessageBoxButtons.OK, MessageBoxIcon.Stop);
                return false;
            }
        }

        /// <summary> build the capture graph for grabber. </summary>
        bool SetupGraph()
        {
            int hr;
            try
            {
                hr = capGraph.SetFiltergraph(graphBuilder);
                if (hr < 0)
                    Marshal.ThrowExceptionForHR(hr);

                hr = graphBuilder.AddFilter(capFilter, "Ds.NET Video Capture Device");
                if (hr < 0)
                    Marshal.ThrowExceptionForHR(hr);

                // Disabled the "Properties" dialogue from showing up everytime the program starts
                //DsUtils.ShowCapPinDialog(capGraph, capFilter, this.Handle);

                AMMediaType media = new AMMediaType();
                media.majorType = MediaType.Video;
                media.subType = MediaSubType.RGB24;
                media.formatType = FormatType.VideoInfo;

                // The following hard codes the video capture size to 640 x 480
                media.formatSize = 0;
                media.sampleSize = 0;

                hr = sampGrabber.SetMediaType(media);
                if (hr < 0)
                    Marshal.ThrowExceptionForHR(hr);

                hr = graphBuilder.AddFilter(baseGrabFlt, "Ds.NET Grabber");
                if (hr < 0)
                    Marshal.ThrowExceptionForHR(hr);

                Guid cat = PinCategory.Preview;
                Guid med = MediaType.Video;
                hr = capGraph.RenderStream(ref cat, ref med, capFilter, null, null); // baseGrabFlt 
                if (hr < 0)
                    Marshal.ThrowExceptionForHR(hr);

                cat = PinCategory.Capture;
                med = MediaType.Video;
                hr = capGraph.RenderStream(ref cat, ref med, capFilter, null, baseGrabFlt); // baseGrabFlt 
                if (hr < 0)
                    Marshal.ThrowExceptionForHR(hr);

                media = new AMMediaType();
                hr = sampGrabber.GetConnectedMediaType(media);
                if (hr < 0)
                    Marshal.ThrowExceptionForHR(hr);
                if ((media.formatType != FormatType.VideoInfo) || (media.formatPtr == IntPtr.Zero))
                    throw new NotSupportedException("Unknown Grabber Media Format");

                videoInfoHeader = (VideoInfoHeader)Marshal.PtrToStructure(media.formatPtr, typeof(VideoInfoHeader));
                Marshal.FreeCoTaskMem(media.formatPtr); media.formatPtr = IntPtr.Zero;

                hr = sampGrabber.SetBufferSamples(false);
                if (hr == 0)
                    hr = sampGrabber.SetOneShot(false);
                if (hr == 0)
                    hr = sampGrabber.SetCallback(null, 0);
                if (hr < 0)
                    Marshal.ThrowExceptionForHR(hr);

                return true;
            }
            catch (Exception ee)
            {
                MessageBox.Show(this, "Could not setup graph\r\n" + ee.Message, "Webcam Signature", MessageBoxButtons.OK, MessageBoxIcon.Stop);
                return false;
            }
        }


        /// <summary> create the used COM components and get the interfaces. </summary>
        bool GetInterfaces()
        {
            Type comType = null;
            object comObj = null;
            try
            {
                comType = Type.GetTypeFromCLSID(Clsid.FilterGraph);
                if (comType == null)
                    throw new NotImplementedException(@"DirectShow FilterGraph not installed/registered!");
                comObj = Activator.CreateInstance(comType);
                graphBuilder = (IGraphBuilder)comObj; comObj = null;

                Guid clsid = Clsid.CaptureGraphBuilder2;
                Guid riid = typeof(ICaptureGraphBuilder2).GUID;
                comObj = DsBugWO.CreateDsInstance(ref clsid, ref riid);
                capGraph = (ICaptureGraphBuilder2)comObj; comObj = null;

                comType = Type.GetTypeFromCLSID(Clsid.SampleGrabber);
                if (comType == null)
                    throw new NotImplementedException(@"DirectShow SampleGrabber not installed/registered!");
                comObj = Activator.CreateInstance(comType);
                sampGrabber = (ISampleGrabber)comObj; comObj = null;

                mediaCtrl = (IMediaControl)graphBuilder;
                videoWin = (IVideoWindow)graphBuilder;
                mediaEvt = (IMediaEventEx)graphBuilder;
                baseGrabFlt = (IBaseFilter)sampGrabber;
                return true;
            }
            catch (Exception ee)
            {
                MessageBox.Show(this, "Could not get interfaces\r\n" + ee.Message, "Webcam Signature", MessageBoxButtons.OK, MessageBoxIcon.Stop);
                return false;
            }
            finally
            {
                if (comObj != null)
                    Marshal.ReleaseComObject(comObj); comObj = null;
            }
        }

        /// <summary> create the user selected capture device. </summary>
        bool CreateCaptureDevice(UCOMIMoniker mon)
        {
            object capObj = null;
            try
            {
                Guid gbf = typeof(IBaseFilter).GUID;
                mon.BindToObject(null, null, ref gbf, out capObj);
                capFilter = (IBaseFilter)capObj; capObj = null;
                return true;
            }
            catch (Exception ee)
            {
                MessageBox.Show(this, "Could not create capture device\r\n" + ee.Message, "Webcam Signature", MessageBoxButtons.OK, MessageBoxIcon.Stop);
                return false;
            }
            finally
            {
                if (capObj != null)
                    Marshal.ReleaseComObject(capObj); capObj = null;
            }

        }

        /// <summary> do cleanup and release DirectShow. </summary>
        void CloseInterfaces()
        {
            int hr;
            try
            {

                if (mediaCtrl != null)
                {
                    hr = mediaCtrl.Stop();
                    mediaCtrl = null;
                }

                if (mediaEvt != null)
                {
                    hr = mediaEvt.SetNotifyWindow(IntPtr.Zero, WM_GRAPHNOTIFY, IntPtr.Zero);
                    mediaEvt = null;
                }

                if (videoWin != null)
                {
                    hr = videoWin.put_Visible(DsHlp.OAFALSE);
                    hr = videoWin.put_Owner(IntPtr.Zero);
                    videoWin = null;
                }

                baseGrabFlt = null;
                if (sampGrabber != null)
                    Marshal.ReleaseComObject(sampGrabber); sampGrabber = null;

                if (capGraph != null)
                    Marshal.ReleaseComObject(capGraph); capGraph = null;

                if (graphBuilder != null)
                    Marshal.ReleaseComObject(graphBuilder); graphBuilder = null;

                if (capFilter != null)
                    Marshal.ReleaseComObject(capFilter); capFilter = null;

                if (capDevices != null)
                {
                    foreach (DsDevice d in capDevices)
                        d.Dispose();
                    capDevices = null;
                }
            }
            catch (Exception)
            { }
        }

        /// <summary> resize preview video window to fill client area. </summary>
        void ResizeVideoWindow()
        {
            if (videoWin != null)
            {
                Rectangle rc = new Rectangle(0, 0, iWidth / 5, iHeight / 5);
                videoWin.SetWindowPosition(0, 0, rc.Right, rc.Bottom);
            }
        }

        /// <summary> override window fn to handle graph events. </summary>
        protected override void WndProc(ref Message m)
        {
            if (m.Msg == WM_GRAPHNOTIFY)
            {
                if (mediaEvt != null)
                    OnGraphNotify();
                return;
            }
            base.WndProc(ref m);
        }

        /// <summary> graph event (WM_GRAPHNOTIFY) handler. </summary>
        void OnGraphNotify()
        {
            DsEvCode code;
            int p1, p2, hr = 0;
            do
            {
                hr = mediaEvt.GetEvent(out code, out p1, out p2, 0);
                if (hr < 0)
                    break;
                hr = mediaEvt.FreeEventParams(code, p1, p2);
            }
            while (hr == 0);
        }

        private void GetAFrame()
        {
            int hr;
            if (savedArray == null)
            {
                int size = videoInfoHeader.BmiHeader.ImageSize;
                if ((size < 1000) || (size > 16000000))
                    return;
                savedArray = new byte[size + 64000];
            }

            captured = false;
            hr = sampGrabber.SetCallback(this, 1);
        }

        /// <summary> capture event, triggered by buffer callback. </summary>
        void OnCaptureDone()
        {
            GCHandle handle;
            try
            {
                int hr;
                if (sampGrabber == null)
                    return;
                hr = sampGrabber.SetCallback(null, 0);

                //return;
                int w = videoInfoHeader.BmiHeader.Width;
                int h = videoInfoHeader.BmiHeader.Height;
                if (((w & 0x03) != 0) || (w < 32) || (w > 4096) || (h < 32) || (h > 4096))
                    return;
                int stride = w * 3;

                handle = GCHandle.Alloc(savedArray, GCHandleType.Pinned);
                int scan0 = (int)handle.AddrOfPinnedObject();
                scan0 += (h - 1) * stride;

                if (DoneProcessing)
                    ProcessPixels = new Bitmap(w, h, -stride, PixelFormat.Format24bppRgb, (IntPtr)scan0);

                handle.Free();
                savedArray = null;
            }
            catch (Exception ee)
            {
                savedArray = null;
                MessageBox.Show(this, "Could not grab picture\r\n" + ee.Message, "Webcam Signature", MessageBoxButtons.OK, MessageBoxIcon.Stop);
            }
        }

        /// <summary> sample callback, NOT USED. </summary>
        int ISampleGrabberCB.SampleCB(double SampleTime, IMediaSample pSample)
        {
            return 0;
        }

        /// <summary> buffer callback, COULD BE FROM FOREIGN THREAD. </summary>
        int ISampleGrabberCB.BufferCB(double SampleTime, IntPtr pBuffer, int BufferLen)
        {
            if (captured || (savedArray == null))
            {
                return 0;
            }

            captured = true;
            bufferedSize = BufferLen;
            if ((pBuffer != IntPtr.Zero) && (BufferLen > 1000) && (BufferLen <= savedArray.Length))
                Marshal.Copy(pBuffer, savedArray, 0, BufferLen);
            this.BeginInvoke(new CaptureDone(this.OnCaptureDone));
            return 0;
        }
        #endregion

        #region Initialize

        public MainForm()
        {
            InitializeComponent();

            // Currently, Webcam Signature is a fixed size Window
            // Webcam Signature 2.0 will allow resizing
            //this.Size = new Size(iWidth + splitContainer1.Panel1.Width + SystemInformation.FrameBorderSize.Width,
            //    iHeight + SystemInformation.FrameBorderSize.Height * 1 + SystemInformation.CaptionHeight);
            this.Size = new Size(800, 600);
            splitContainer1.SplitterDistance = 150;
            //Signature.Width = iWidth;
            //Signature.Height = iHeight;
            //DrawingWidth = SystemInformation.PrimaryMonitorSize.Width;
            //DrawingHeight = SystemInformation.PrimaryMonitorSize.Height;
            DrawingWidth = 800;
            DrawingHeight = 600;
            Signature.Width = DrawingWidth;
            Signature.Height = DrawingHeight;

            // Load the color of signature and background from the UserSettings
            ColorConverter colconv = new ColorConverter();
            SignatureColorSystem = (Color)colconv.ConvertFromString(UserSettings.Default.LineColor);
            SignatureColor.blue = (byte)SignatureColorSystem.B;
            SignatureColor.green = (byte)SignatureColorSystem.G;
            SignatureColor.red = (byte)SignatureColorSystem.R;

            backColorSystem = (Color)colconv.ConvertFromString(UserSettings.Default.BackColor);
            backColor.blue = (byte)backColorSystem.B;
            backColor.green = (byte)backColorSystem.G;
            backColor.red = (byte)backColorSystem.R;

            // Create the special color selection buttons
            PenColorButton.Size = new Size(57, 25);
            PenColorButton.Location = new Point(8, 15);
            PenColorButton.Click += new EventHandler(PenColorButton_Click);
            PenColorButton.BackColor = SystemColors.Control;
            PenColorButton.CenterColor = SignatureColorSystem;
            PenColorButton.FlatStyle = FlatStyle.Popup;
            groupBoxPen.Controls.Add(PenColorButton);
 
            BackgroundColorButton.Size = new Size(57, 25);
            BackgroundColorButton.Location = new Point(8, 15);
            BackgroundColorButton.Click += new EventHandler(BackgroundColorButton_Click);
            BackgroundColorButton.BackColor = SystemColors.Control;
            BackgroundColorButton.CenterColor = backColorSystem;
            BackgroundColorButton.FlatStyle = FlatStyle.Popup;
            groupBoxBackground.Controls.Add(BackgroundColorButton);

            if (UserSettings.Default.TransparentOnSave)
                MakeBackTransparent.Checked = true;
            else
                MakeBackTransparent.Checked = false;

            WebCamBitmap = null;
            Signature.BringToFront();
            ShowPointLocation.BringToFront();
            SignaturePixels = new System.Drawing.Bitmap(DrawingWidth, DrawingHeight);
            InitiatilizeSignaturePixels(ref SignaturePixels);

            ProcessPixels = new System.Drawing.Bitmap(iWidth, iHeight);

            Signature.Image = SignaturePixels;

            LoadTrackColor();
            UpdateTrackColorSample();

            SignaturePoints.Clear();
            SignaturePoints.Add(null);
            Undo.Clear();
            Redo.Clear();
            Undo.Add(SignaturePoints);

            // Tooltips are used sparingly
            toolTipTransparentOnSave.SetToolTip(this.MakeBackTransparent, 
                "Transparent backgrounds are NOT compatible with MS Paint/Photoshop.\n"+
                "Works with GIMP/MS Office 2007/PDF's.\n" + "The clipboard does not support transparent backgrounds.");
            HelpAboutToolTip.SetToolTip(this.HelpAbout, "Help & About");

            // A timer is used to control the analysis of the video from the webcam
            timer1.Interval = (int)(1.0 / (double)FrameRate * 1000.0);
            timer1.Enabled = true;
            timer1.Start();

            isKeyDown = false;
            DoneProcessing = true;
            FormActivated = true;
            FineTuneTC = null;

            PenThickness = UserSettings.Default.LineThickness;
            PenSizeDropBox.Text = UserSettings.Default.LineThickness.ToString();

            ShowPointLocation.Location = new Point(-100, -100);
            InitializeBackgroundAverage();
            RefreshBackFilteringParams();

            WebcamMouseMode = false;
            WebcamMouseCB.Checked = false;
        }

        private void InitializeBackgroundAverage()
        {
            BackgroundAverage = new double[iWidth, iHeight, 3];
            for (int i = 0; i < iWidth; i++)
            {
                for (int j = 0; j < iHeight; j++)
                {
                    for (int k = 0; k < 3; k++)
                    {
                        BackgroundAverage[i, j, k] = 0.0;
                    }
                }
            }
        }

        private void MainForm_Shown(object sender, EventArgs e)
        {
            //if (UserSettings.Default.ActivateWebcam)
            //{
                InitializeCamera();
            //}
        }

        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            CloseInterfaces();
        }

        // Show the colors that the program is tracking
        public void UpdateTrackColorSample()
        {
            if (UserSettings.Default.TrackRed)
            {
                this.TrackRed.Checked = true;
                this.TrackGreen.Checked = false;
                this.TrackBlue.Checked = false;
                SampleTrackColor.BackColor = Color.FromArgb(RedH, GreenH, BlueH);
                SampleTrackColorLow.BackColor = Color.FromArgb(RedL, GreenL, BlueL);
            }
            else if (UserSettings.Default.TrackGreen)
            {
                this.TrackRed.Checked = false;
                this.TrackGreen.Checked = true;
                this.TrackBlue.Checked = false;
                SampleTrackColor.BackColor = Color.FromArgb(RedH, GreenH, BlueH);
                SampleTrackColorLow.BackColor = Color.FromArgb(RedL, GreenL, BlueL);
            }
            else
            {
                this.TrackRed.Checked = false;
                this.TrackGreen.Checked = false;
                this.TrackBlue.Checked = true;
                SampleTrackColor.BackColor = Color.FromArgb(RedH, GreenH, BlueH);
                SampleTrackColorLow.BackColor = Color.FromArgb(RedL, GreenL, BlueL);
            }
        }

        // Get the colors to track from UserSettings
        public void LoadTrackColor()
        {
            if (UserSettings.Default.TrackRed)
            {
                RedL = UserSettings.Default.TRLowRed;
                RedH = UserSettings.Default.TRHighRed;
                GreenL = UserSettings.Default.TRLowGreen;
                GreenH = UserSettings.Default.TRHighGreen;
                BlueL = UserSettings.Default.TRLowBlue;
                BlueH = UserSettings.Default.TRHighBlue;
            }
            else if (UserSettings.Default.TrackGreen)
            {
                RedL = UserSettings.Default.TGLowRed;
                RedH = UserSettings.Default.TGHighRed;
                GreenL = UserSettings.Default.TGLowGreen;
                GreenH = UserSettings.Default.TGHighGreen;
                BlueL = UserSettings.Default.TGLowBlue;
                BlueH = UserSettings.Default.TGHighBlue;
            }
            else
            {
                RedL = UserSettings.Default.TBLowRed;
                RedH = UserSettings.Default.TBHighRed;
                GreenL = UserSettings.Default.TBLowGreen;
                GreenH = UserSettings.Default.TBHighGreen;
                BlueL = UserSettings.Default.TBLowBlue;
                BlueH = UserSettings.Default.TBHighBlue;
            }
        }

        #endregion

        #region Core Code

        // timer1 controls when the program draws pixels to the screen
        private void timer1_Tick(object sender, EventArgs e)
        {
            if (DoneProcessing && FormActivated)
                DrawPixels();
        }

        // This is the main function that coordinates the drawing of the Signature pixels
        // It also filters the input to prevent jittering of the image.
        // This is reason why the signatures come out smooth.
        private void DrawPixels()
        {
            DoneProcessing = false;
            try
            {
                Point pt = new Point(0, 0);
                Point ScaledPoint = new Point(0, 0);

                if (ProcessPixels != null)
                {
                    GetAFrame();

                    if (GetPoint(ref pt, ref ProcessPixels, RedL, RedH, GreenL, GreenH, BlueL, BlueH) == 1)
                    {
                        if (WebcamMouseMode == false)
                        {
                            ScaledPoint.X = pt.X - (int)((double)iWidth * 0.1 / 2.0);
                            ScaledPoint.Y = pt.Y - (int)((double)iHeight * 0.1 / 2.0);
                            //ScaledPoint = TranslateWebcamPTtoScreen(iWidth, iHeight,
                            //    (int)((double)iWidth * 0.9), (int)((double)iHeight * 0.9), ScaledPoint);
                            ScaledPoint = TranslateWebcamPTtoScreen(SystemInformation.PrimaryMonitorSize.Width, SystemInformation.PrimaryMonitorSize.Height,
                                (int)((double)iWidth * 0.9), (int)((double)iHeight * 0.9), ScaledPoint);

                            if (isKeyDown && !(SignaturePoints.Count > 0 && SignaturePoints[SignaturePoints.Count - 1] != null && ScaledPoint == (Point)SignaturePoints[SignaturePoints.Count - 1]))
                            {
                                SignaturePoints.Add(ScaledPoint);

                                // Do a rough representation of the signature while the user draws it
                                if (SignaturePoints.Count >= 2 && SignaturePoints[SignaturePoints.Count - 2] != null &&
                                    SignaturePoints[SignaturePoints.Count - 1] != null)
                                {
                                    DrawLine((Point)SignaturePoints[SignaturePoints.Count - 2], (Point)SignaturePoints[SignaturePoints.Count - 1],
                                        ref SignaturePixels, true);
                                }
                                else if (SignaturePoints.Count >= 1 && SignaturePoints[SignaturePoints.Count - 1] != null)
                                {
                                    DrawAPoint(ScaledPoint, ref SignaturePixels, true);
                                }
                            }

                            ShowPointLocation.Location = new Point(ScaledPoint.X, ScaledPoint.Y);
                        }
                        else
                        {
                            ScaledPoint.X = pt.X - (int)((double)iWidth * 0.1 / 2.0);
                            ScaledPoint.Y = pt.Y - (int)((double)iHeight * 0.1 / 2.0);
                            Cursor.Position = TranslateWebcamPTtoScreen(SystemInformation.PrimaryMonitorSize.Width, SystemInformation.PrimaryMonitorSize.Height,
                                (int)((double)iWidth * 0.9), (int)((double)iHeight * 0.9), ScaledPoint);
                        }
                    }
                }

            }
            catch (Exception)
            {
                // Do nothing
            }

            ProcessPixels = null;
            GC.Collect();
            DoneProcessing = true;
        }

        // Set the background color of the Signature
        private void InitiatilizeSignaturePixels(ref Bitmap NewBitmap)
        {
            UnsafeBitmap WebCamBitmap = new UnsafeBitmap(ref NewBitmap, true);
            WebCamBitmap.LockBitmap();

            for (int i = 0; i < WebCamBitmap.Bitmap.Width; i++)
            {
                for (int j = 0; j < WebCamBitmap.Bitmap.Height; j++)
                {
                    WebCamBitmap.SetPixel(i, j, backColor);
                }
            }
            WebCamBitmap.UnlockBitmap();
        }

        // Draw a circular point with a radius of PenThickness / 2 
        public void DrawAPoint(Point pt, ref Bitmap image, bool RefreshRegion)
        {
            UnsafeBitmap WebCamBitmap = new UnsafeBitmap(ref image, true);
            WebCamBitmap.LockBitmap();

            int left = pt.X - (int)(PenThickness / (float)2.0) + 1;
            if (left < 0)
                left = 0;

            int top = pt.Y - (int)(PenThickness / (float)2.0) + 1;
            if (top < 0)
                top = 0;

            int right = pt.X + (int)(PenThickness / (float)2.0) - 1;
            if (right > image.Width)
                right = image.Width;

            int bottom = pt.Y + (int)(PenThickness / (float)2.0) - 1;
            if (bottom > image.Height)
                bottom = image.Height;

            double radius = (double)(bottom - top) / 2.0;

            for (int i = left; i < right; i++)
            {
                for (int j = top; j < bottom; j++)
                {
                    if (Math.Sqrt(Math.Pow(Math.Abs(i - pt.X), 2.0) + Math.Pow(Math.Abs(j - pt.Y), 2.0)) <= radius)
                    {
                        WebCamBitmap.SetPixel(i, j, SignatureColor);
                    }
                }
            }
            WebCamBitmap.UnlockBitmap();

            // Redraw only the region necessary to increase speed
            if (RefreshRegion)
            {
                Rectangle tempRect = new Rectangle(left, top, right, bottom);
                Region r = new Region(tempRect);
                Signature.Invalidate(r);
                Signature.Update();
            }
        }

        // Draws a line with a start and end point on a given image.
        private void DrawLine(Point start, Point end, ref Bitmap image, bool RefreshRegion)
        {
            if (Math.Sqrt(Math.Pow(Math.Abs(start.X - end.X), 2.0) + Math.Pow(Math.Abs(start.Y - end.Y), 2.0)) <= (double)PenThickness)
            {
                Point middle = new Point((start.X + end.X) / 2, (start.Y + end.Y) / 2);
                DrawAPoint(middle, ref image, RefreshRegion);
            }
            else
            {
                Graphics g = null;
                g = Graphics.FromImage((Image)image);
                Pen pen = new Pen(SignatureColorSystem, PenThickness);
                pen.MiterLimit = 0.0f;
                pen.StartCap = System.Drawing.Drawing2D.LineCap.Round;
                pen.EndCap = System.Drawing.Drawing2D.LineCap.Round;
                g.DrawLine(pen,start.X, start.Y, end.X, end.Y);
                pen.Dispose();
                g.Dispose();

                // Refresh only the region drawn to speed up the program
                // instead of using Signature.Refresh() to update the whole image
                if (RefreshRegion)
                {
                    int tempLeft, tempTop, tempRight, tempBottom;
                    if (start.X < end.X)
                    {
                        tempLeft = start.X;
                        tempRight = end.X;
                    }
                    else
                    {
                        tempLeft = end.X;
                        tempRight = start.X;
                    }

                    if (start.Y < end.Y)
                    {
                        tempTop = start.Y;
                        tempBottom = end.Y;
                    }
                    else
                    {
                        tempTop = end.Y;
                        tempBottom = start.Y;
                    }

                    Rectangle tempRect = new Rectangle(tempLeft, tempTop, tempRight, tempBottom);
                    Region r = new Region(tempRect);
                    Signature.Invalidate(r);
                    Signature.Update();
                }
            }
        }

        // This button is hidden on the user screen.
        // It can be shown to allow the user to do manual smoothing with some coding.
        private void ApplySmoothing_MouseClick(object sender, MouseEventArgs e)
        {
            InitiatilizeSignaturePixels(ref SignaturePixels);
            RedrawPixels(ref SignaturePixels);
        }

        // Called by ApplySmoothing_MouseClick to redraw the signature area
        private void RedrawPixels(ref Bitmap TargetBitmap)
        {
            int lastIndex = 0;
            while (lastIndex < SignaturePoints.Count - 1)
            {
                int null1 = SignaturePoints.IndexOf(null, lastIndex);
                if (null1 < 0 || null1 == SignaturePoints.Count - 1)
                {
                    break;
                }

                int null2 = SignaturePoints.IndexOf(null, null1 + 1);
                lastIndex = null2;
                if (null2 < 0)
                {
                    break;
                }

                int curvePointsSize = null2 - null1 - 1;

                if (curvePointsSize >= 3)
                {
                    Point[] curvePoints;
                    curvePoints = new Point[curvePointsSize];

                    int tempCounter = 0;
                    for (int i = null1 + 1; i < null2; i++)
                    {
                        curvePoints[tempCounter] = (Point)SignaturePoints[i];
                        tempCounter++;
                    }

                    ApplySmoothingFunction(curvePoints, ref TargetBitmap);
                }
                else if (curvePointsSize == 2)
                {
                    DrawLine((Point)SignaturePoints[null1 + 1], (Point)SignaturePoints[null1 + 2], ref TargetBitmap, false);
                }
                else if (curvePointsSize == 1)
                {
                    DrawAPoint((Point)SignaturePoints[null1 + 1], ref TargetBitmap, false);
                }
            }
            Signature.Refresh();
        }

        // ApplySmoothingFunction creates smooth cardinal curves for an array of points
        private void ApplySmoothingFunction(Point[] curvePoints, ref Image image)
        {
            Graphics g = null;
            g = Graphics.FromImage(image);
            Pen pen = new Pen(SignatureColorSystem, PenThickness);
            pen.MiterLimit = 0.0f;
            pen.StartCap = System.Drawing.Drawing2D.LineCap.Round;
            pen.EndCap = System.Drawing.Drawing2D.LineCap.Round;
            g.DrawCurve(pen, curvePoints, (float)0.5);

            pen.Dispose();
            g.Dispose();
        }

        private void ApplySmoothingFunction(Point[] curvePoints, ref Bitmap image)
        {
            Graphics g = null;
            g = Graphics.FromImage((Image)image);
            Pen pen = new Pen(SignatureColorSystem, PenThickness);
            pen.MiterLimit = 0.0f;
            pen.StartCap = System.Drawing.Drawing2D.LineCap.Round;
            pen.EndCap = System.Drawing.Drawing2D.LineCap.Round;
            g.DrawCurve(pen, curvePoints, (float)0.5);

            pen.Dispose();
            g.Dispose();
        }

        // MovingAverage is not currently used
        private void MovingAverage(ref Point[] curvePoints)
        {
            long sumX = 0;
            long sumY = 0;
            long numPoints = 0;

            for (int i = 0; i < curvePoints.Length; i++)
            {
                Point pt = curvePoints[i];
                sumX = sumX + (long)pt.X;
                sumY = sumY + (long)pt.Y;
                numPoints++;

                pt.X = (int)((double)sumX / (double)numPoints);
                pt.Y = (int)((double)sumY / (double)numPoints);

                curvePoints[i] = pt;
            }
        }


        // GetPoint takes a color bitmap image and filters out everything but the given RBG color range.
        // It returns 1 if it finds a point, 0 other wise.
        // The point it finds is placed in CurrentPoint.
        // This is the greatest bottleneck of the program.
        private int GetPoint(ref Point CurrentPoint, ref Bitmap RawPixels, int RedLow, int RedHigh, int GreenLow, int GreenHigh, int BlueLow, int BlueHigh)
        {
            int SumX = 0;
            int SumY = 0;
            int NumPoints = 0;
            int neighbors = 0;

            // The UnsafeBitmap greatly speeds up the processing time.
            WebCamBitmap = new UnsafeBitmap(RawPixels);
            WebCamBitmap.LockBitmap();

            // We don't actually look at every pixel because that consumes too much processing time
            // We look at every other one
            for (int i = 0; i < WebCamBitmap.Bitmap.Width; i += 2)
            {
                for (int j = 0; j < WebCamBitmap.Bitmap.Height; j += 2)
                {
                    PixelData pixel = WebCamBitmap.GetPixel(i, j);

                    // Update the background averages
                    // Can be used for static background filtering or
                    // it can be used for dynamic background filtering
                    // where it will constantly update itself
                    if (NoBackgroundFilter == false)
                    {
                        if (NumBackgroundFramesCaptured < RequiredNumBackgroundFrames)
                        {
                            if (NumBackgroundFramesCaptured == 0)
                            {
                                //red
                                BackgroundAverage[i, j, 0] = (double)pixel.red;
                                //green
                                BackgroundAverage[i, j, 1] = (double)pixel.green;
                                //blue
                                BackgroundAverage[i, j, 2] = (double)pixel.blue;
                            }
                            else
                            {
                                //red
                                BackgroundAverage[i, j, 0] = (((double)BackgroundAverage[i, j, 0] * (double)(NumBackgroundFramesCaptured) + (double)pixel.red)
                                    / (double)(NumBackgroundFramesCaptured + 1));
                                //green
                                BackgroundAverage[i, j, 1] = (((double)BackgroundAverage[i, j, 1] * (double)(NumBackgroundFramesCaptured) + (double)pixel.green)
                                    / (double)(NumBackgroundFramesCaptured + 1));
                                //blue
                                BackgroundAverage[i, j, 2] = (((double)BackgroundAverage[i, j, 2] * (double)(NumBackgroundFramesCaptured) + (double)pixel.blue)
                                    / (double)(NumBackgroundFramesCaptured + 1));
                            }
                        }
                        else if (UseBackgroundDynamicFilter == true)
                        {
                            //red
                            BackgroundAverage[i, j, 0] = (((double)BackgroundAverage[i, j, 0] * (double)(NumBackgroundFramesCaptured) + (double)pixel.red)
                                / (double)(NumBackgroundFramesCaptured + 1));
                            //green
                            BackgroundAverage[i, j, 1] = (((double)BackgroundAverage[i, j, 1] * (double)(NumBackgroundFramesCaptured) + (double)pixel.green)
                                / (double)(NumBackgroundFramesCaptured + 1));
                            //blue
                            BackgroundAverage[i, j, 2] = (((double)BackgroundAverage[i, j, 2] * (double)(NumBackgroundFramesCaptured) + (double)pixel.blue)
                                / (double)(NumBackgroundFramesCaptured + 1));
                        }
                    }


                    if ((Math.Abs(BackgroundAverage[i, j, 0] - (int)pixel.red) > BackgroundDiffThreshold ||
                        Math.Abs(BackgroundAverage[i, j, 1] - (int)pixel.green) > BackgroundDiffThreshold ||
                        Math.Abs(BackgroundAverage[i, j, 2] - (int)pixel.blue) > BackgroundDiffThreshold || NoBackgroundFilter) &&
                        pixel.red >= (byte)RedLow && pixel.red <= (byte)RedHigh && pixel.green >= (byte)GreenLow &&
                        pixel.green <= (byte)GreenHigh && pixel.blue >= (byte)BlueLow && pixel.blue <= (byte)BlueHigh)
                    {
                        // We now know that the pixel is a valid color
                        // We now need to check if there is a neighboring pixel of a valid color
                        // Purpose is to filter out errant pixels
                        // Right now, it is coded to look in a region of 100 pixels for
                        // each valid pixel found to see if there is a neighbor
                        int ktemp = i - 4;
                        if (ktemp < 0)
                            ktemp = 0;

                        int khigh = i + 4;
                        if (khigh > WebCamBitmap.Bitmap.Width - 1)
                        {
                            khigh = WebCamBitmap.Bitmap.Width - 1;
                        }

                        int ltemp = j - 4;
                        if (ltemp < 0)
                            ltemp = 0;

                        int lhigh = j + 4;
                        if (lhigh > WebCamBitmap.Bitmap.Height - 1)
                        {
                            lhigh = WebCamBitmap.Bitmap.Height - 1;
                        }

                        neighbors = 0;

                        for (int k = ktemp; k <= khigh; k++)
                        {
                            for (int l = ltemp; l <= lhigh; l++)
                            {
                                PixelData NeighborPixel = WebCamBitmap.GetPixel(k, l);
                                if (NeighborPixel.red >= (byte)RedLow && NeighborPixel.red <= (byte)RedHigh && NeighborPixel.green >= (byte)GreenLow &&
                                    NeighborPixel.green <= (byte)GreenHigh && NeighborPixel.blue >= (byte)BlueLow && NeighborPixel.blue <= (byte)BlueHigh)
                                {
                                    neighbors++;
                                }
                            }
                            if (neighbors >= 1)
                                break;
                        }

                        if (neighbors >= 1)
                        {
                            NumPoints++;
                            SumX += WebCamBitmap.Bitmap.Width - i;
                            SumY += j;
                        }
                    }
                }
            }
            WebCamBitmap.UnlockBitmap();
            WebCamBitmap.Dispose();

            if (NoBackgroundFilter == false)
            {
                if (NumBackgroundFramesCaptured < RequiredNumBackgroundFrames)
                {
                    NumBackgroundFramesCaptured++;
                    if (NumBackgroundFramesCaptured >= NumBackgroundFramesRequiredCalibration)
                        NumBackgroundFramesCaptured = RequiredNumBackgroundFrames;
                }
            }

            if (NumPoints >= 1)
            {
                // The point we return is simply the average of the points seen
                // No need to over complicate this by calculating the centroid
                CurrentPoint = new Point(SumX / NumPoints, SumY / NumPoints);
                return 1;
            }
            else
            {
                return 0;
            }
        }

        private Point TranslateWebcamPTtoScreen(int ScreenWidth, int ScreenHeight, int WebcamWidth, int WebcamHeight, Point CurrPoint)
        {
            Point ScreenPoint = new Point();

            ScreenPoint.X = (int)((double)CurrPoint.X / (double)WebcamWidth * (double)ScreenWidth);
            ScreenPoint.Y = (int)((double)CurrPoint.Y / (double)WebcamHeight * (double)ScreenHeight);

            if (ScreenPoint.X < 0)
                ScreenPoint.X = 0;
            if (ScreenPoint.X > ScreenWidth)
                ScreenPoint.X = ScreenWidth;
            if (ScreenPoint.Y < 0)
                ScreenPoint.Y = 0;
            if (ScreenPoint.Y > ScreenHeight)
                ScreenPoint.Y = ScreenHeight;

            //Console.WriteLine(CurrPoint + "  " + ScreenPoint);
            return ScreenPoint;
        }

        public void RefreshBackFilteringParams()
        {
            RequiredNumBackgroundFrames = (int)(UserSettings.Default.BackgroundRefreshTime * (double)FrameRate);
            NumBackgroundFramesCaptured = 0;
            BackgroundDiffThreshold = UserSettings.Default.BackgroundDiffThreshold;
            BackgroundCalibrateTime = UserSettings.Default.BackgroundCalibrateTime;
            NumBackgroundFramesRequiredCalibration = (int)(UserSettings.Default.BackgroundCalibrateTime * (double)FrameRate);
            NoBackgroundFilter = UserSettings.Default.NoBackgroundFilter;
            UseBackgroundStaticFilter = UserSettings.Default.UseStaticBackFilter;
            UseBackgroundDynamicFilter = UserSettings.Default.UseDynamicBackFilter;
            //InitializeBackgroundAverage();
        }

        // Creates a Bitmap from the array of values for the background average
        public Bitmap GetStaticBackBitmap()
        {
            Bitmap StaticBackground = new Bitmap(iWidth, iHeight);

            if (mediaCtrl != null)
            {
                int state1, state2, state3;
                state2 = 500;
                state1 = mediaCtrl.GetState(state2, out state3);

                if (state3 == 2)
                {
                    UnsafeBitmap WebCamBitmap = new UnsafeBitmap(ref StaticBackground, true);
                    WebCamBitmap.LockBitmap();

                    for (int i = 0; i < WebCamBitmap.Bitmap.Width; i += 2)
                    {
                        for (int j = 0; j < WebCamBitmap.Bitmap.Height; j += 2)
                        {
                            PixelData currPixel = new PixelData();
                            currPixel.red = (byte)BackgroundAverage[i, j, 0];
                            currPixel.green = (byte)BackgroundAverage[i, j, 1];
                            currPixel.blue = (byte)BackgroundAverage[i, j, 2];
                            WebCamBitmap.SetPixel(i, j, currPixel);
                        }
                    }
                    WebCamBitmap.UnlockBitmap();
                }
            }

            GC.Collect();
            return StaticBackground;
        }

        // Returns a full frame captured from the webcam
        // This is used to pass a frame to FineTuneTrackColorsForm
        public Bitmap GetFullFrameBitmap()
        {
            Bitmap ReturnBitmap = new Bitmap(iWidth, iHeight);

            try
            {
                if (mediaCtrl != null)
                {
                    int state1, state2, state3;
                    state2 = 500;
                    state1 = mediaCtrl.GetState(state2, out state3);

                    if (state3 == 2)
                    {
                        DoneProcessing = false;
                        try
                        {
                            if (ProcessPixels != null)
                            {
                                GetAFrame();
                            }
                        }
                        catch (Exception)
                        {
                        }

                        //UnsafeBitmap WebCamBitmap = new UnsafeBitmap(ProcessPixels);
                        //WebCamBitmap.LockBitmap();
                        //UnsafeBitmap UnsafeReturnBitmap = new UnsafeBitmap(ref ReturnBitmap, true);
                        //UnsafeReturnBitmap.LockBitmap();

                        //for (int i = 0; i < WebCamBitmap.Bitmap.Width; i++)
                        //{
                        //    for (int j = 0; j < WebCamBitmap.Bitmap.Height; j++)
                        //    {
                        //        UnsafeReturnBitmap.SetPixel(i, j, WebCamBitmap.GetPixel(i, j));
                        //    }
                        //}
                        //WebCamBitmap.UnlockBitmap();
                        //UnsafeReturnBitmap.UnlockBitmap();
                        if (ProcessPixels != null)
                            ReturnBitmap = new Bitmap(ProcessPixels);

                        ProcessPixels = null;
                        DoneProcessing = true;
                    }
                }
            }
            catch (Exception ee)
            {
                Console.WriteLine("Error render frame. " + ee);
            }

            GC.Collect();
            return ReturnBitmap;
        }

        public int RefreshingBackgroundFilterProgress()
        {
            if (mediaCtrl != null)
            {
                int state1, state2, state3;
                state2 = 500;
                state1 = mediaCtrl.GetState(state2, out state3);

                if (state3 == 2)
                {
                    if (NoBackgroundFilter == false)
                    {
                        if (NumBackgroundFramesCaptured < RequiredNumBackgroundFrames)
                        {
                            return (int)((double)NumBackgroundFramesCaptured / (double)RequiredNumBackgroundFrames * (100.0));
                        }
                        else
                        {
                            return 100;
                        }
                    }
                    else
                    {
                        return 100;
                    }
                }
                else
                {
                    return 100;
                }
            }
            else
            {
                return 100;
            }
        }

        #endregion

        #region Change Tracking Colors

        private void TrackRed_CheckedChanged(object sender, EventArgs e)
        {
            if (this.TrackRed.Checked == true)
            {
                UserSettings.Default.TrackRed = true;
                UserSettings.Default.TrackGreen = false;
                UserSettings.Default.TrackBlue = false;
                UserSettings.Default.Save();
                LoadTrackColor();
                UpdateTrackColorSample();

                if (FineTuneTC != null)
                {
                    FineTuneTC.LoadSettings();
                }
            }
        }

        private void TrackGreen_CheckedChanged(object sender, EventArgs e)
        {
            if (this.TrackGreen.Checked == true)
            {
                UserSettings.Default.TrackRed = false;
                UserSettings.Default.TrackGreen = true;
                UserSettings.Default.TrackBlue = false;
                UserSettings.Default.Save();
                LoadTrackColor();
                UpdateTrackColorSample();

                if (FineTuneTC != null)
                {
                    FineTuneTC.LoadSettings();
                }
            }
        }

        private void TrackBlue_CheckedChanged(object sender, EventArgs e)
        {
            if (this.TrackBlue.Checked == true)
            {
                UserSettings.Default.TrackRed = false;
                UserSettings.Default.TrackGreen = false;
                UserSettings.Default.TrackBlue = true;
                UserSettings.Default.Save();
                LoadTrackColor();
                UpdateTrackColorSample();

                if (FineTuneTC != null)
                {
                    FineTuneTC.LoadSettings();
                }
            }
        }

        private void FineTuneTrackColors_MouseClick(object sender, MouseEventArgs e)
        {
            if (FineTuneTC == null)
            {
                FineTuneTC = new FineTuneTrackColorsForm(this);
                FormActivated = true;
            }
            else
            {
                FineTuneTC.Show();
                FineTuneTC.BringToFront();
            }
        }

        #endregion

        #region File Processing

        private void CopytoClipboard_MouseClick(object sender, MouseEventArgs e)
        {
            Clipboard.SetImage((Image)SignaturePixels);
        }

        // Save the signature to an image file
        private void saveFileDialog1_FileOk(object sender, CancelEventArgs e)
        {
            Image tempImage;

            if (UserSettings.Default.TransparentOnSave == true)
            {
                Bitmap transparentVersion = new Bitmap((Image)SignaturePixels);
                for (int i = 0; i < transparentVersion.Width; i++)
                {
                    for (int j = 0; j < transparentVersion.Height; j++)
                    {
                        transparentVersion.SetPixel(i, j, SignatureColorSystem);
                    }
                }

                transparentVersion.MakeTransparent(SignatureColorSystem);
                RedrawPixels(ref transparentVersion);
                tempImage = (Image)transparentVersion;
                tempImage.Save(saveFileDialog1.FileName);
            }
            else
            {
                tempImage = (Image)SignaturePixels;
                tempImage.Save(saveFileDialog1.FileName);
            }
        }

        private void SavetoFile_MouseClick(object sender, MouseEventArgs e)
        {
            saveFileDialog1.ShowDialog();
        }

        // Save the signature data points to a .wsd file
        private void SaveDataPoints_MouseClick(object sender, MouseEventArgs e)
        {
            SaveDataPointsDialog.ShowDialog();
        }

        private void SaveDataPointsDialog_FileOk(object sender, CancelEventArgs e)
        {
            try
            {
                TextWriter tw = new StreamWriter(SaveDataPointsDialog.FileName);

                tw.WriteLine("[Signature Points Start]");
                for (int i = 0; i <= SignaturePoints.Count - 1; i++)
                {
                    if (SignaturePoints[i] == null)
                        tw.WriteLine("null");
                    else
                        tw.WriteLine(SignaturePoints[i].ToString());
                }
                tw.WriteLine("[Signature Points End]");
                tw.Close();
            }
            catch (Exception ee)
            {
                MessageBox.Show(this, "Could not save to file. File access error.", "Webcam Signature", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        // Load signature data points from a .wsd file
        private void LoadDataPoints_MouseClick(object sender, MouseEventArgs e)
        {
            OpenDataPointsDialog.ShowDialog();
        }

        private void OpenDataPointsDialog_FileOk(object sender, CancelEventArgs e)
        {
            try
            {
                ArrayList SigDataPoints = new ArrayList();
                TextReader tr = new StreamReader(OpenDataPointsDialog.FileName);
                while (true)
                {
                    String dataline = tr.ReadLine();
                    if (dataline == null)
                    {
                        break;
                    }
                    else if (dataline == "[Signature Points Start]")
                    {
                        String SigPoint = tr.ReadLine();
                        if (SigPoint == null)
                            break;

                        while (SigPoint != "[Signature Points End]")
                        {
                            if (SigPoint == "null")
                            {
                                SigDataPoints.Add(null);
                                SigPoint = tr.ReadLine();
                            }
                            else
                            {
                                Point pt = new Point(0, 0);
                                pt.X = int.Parse(SigPoint.Substring(SigPoint.IndexOf("X=") + 2, SigPoint.IndexOf(",") - (SigPoint.IndexOf("X=") + 2)));
                                pt.Y = int.Parse(SigPoint.Substring(SigPoint.IndexOf("Y=") + 2, SigPoint.IndexOf("}") - (SigPoint.IndexOf("Y=") + 2)));
                                SigDataPoints.Add(pt);
                                SigPoint = tr.ReadLine();
                            }
                        }
                    }
                }
                tr.Close();

                SignaturePoints.Clear();
                SignaturePoints = SigDataPoints;
                int begin = 1;

                while (begin < SigDataPoints.Count)
                {
                    ArrayList tempList2 = new ArrayList();
                    begin = SigDataPoints.IndexOf(null, begin);

                    if (begin < 0)
                        break;

                    for (int i = 0; i <= begin; i++)
                    {
                        tempList2.Add(SigDataPoints[i]);
                    }

                    SetUndo(tempList2);

                    begin += 1;
                }
                InitiatilizeSignaturePixels(ref SignaturePixels);
                RedrawPixels(ref SignaturePixels);
            }
            catch (Exception ee)
            {
                MessageBox.Show(this, "Could not open file / invalid file." + ee, "Webcam Signature", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        #endregion

        #region Undo/Redo Functions

        private void buttonUndo_MouseClick(object sender, MouseEventArgs e)
        {
            if (Undo.Count > 1 && isKeyDown == false)
            {
                SetRedo(SignaturePoints);
                ArrayList tempUndoPoint = new ArrayList((ArrayList)Undo[Undo.Count - 2]);
                if (tempUndoPoint.Count == 0)
                    tempUndoPoint.Add(null);
                SignaturePoints.Clear();
                SignaturePoints = tempUndoPoint;
                Undo.RemoveAt(Undo.Count - 1);
                InitiatilizeSignaturePixels(ref SignaturePixels);
                RedrawPixels(ref SignaturePixels);
                UpdateUndoRedoButtons();
            }
        }

        private void buttonRedo_MouseClick(object sender, MouseEventArgs e)
        {
            if (Redo.Count > 0 && isKeyDown == false)
            {
                ArrayList tempRedoPoint = new ArrayList((ArrayList)Redo[Redo.Count - 1]);
                if (tempRedoPoint.Count == 0)
                    tempRedoPoint.Add(null);
                SignaturePoints.Clear();
                SignaturePoints = tempRedoPoint;
                SetUndo(SignaturePoints);
                Redo.RemoveAt(Redo.Count - 1);
                InitiatilizeSignaturePixels(ref SignaturePixels);
                RedrawPixels(ref SignaturePixels);
                UpdateUndoRedoButtons();
            }
        }

        private void UpdateUndoRedoButtons()
        {
            if (Undo.Count > 1)
                buttonUndo.Enabled = true;
            else
                buttonUndo.Enabled = false;

            if (Redo.Count > 0)
                buttonRedo.Enabled = true;
            else
                buttonRedo.Enabled = false;

            buttonUndo.Text = "<- " + (Undo.Count - 1);
            buttonRedo.Text = Redo.Count + " ->";
        }

        private void SetUndo(ArrayList DataPoints)
        {
            while (Undo.Count > 200)
            {
                Undo.RemoveAt(0);
            }
            ArrayList tempList = new ArrayList(DataPoints);
            Undo.Add(tempList);
            UpdateUndoRedoButtons();
        }

        private void SetRedo(ArrayList DataPoints)
        {
            while (Redo.Count > 200)
            {
                Redo.RemoveAt(0);
            }
            ArrayList tempList = new ArrayList(DataPoints);
            Redo.Add(tempList);
            UpdateUndoRedoButtons();
        }

        #endregion

        #region User Actions

        private void MainForm_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode.ToString() != "Return" && PenSizeDropBox.Focused == false &&
                (e.KeyCode.ToString() == "ShiftKey" || e.KeyCode.ToString() == "ControlKey"))
            {
                isKeyDown = true;
                e.Handled = true;
            }
        }

        private void MainForm_KeyUp(object sender, KeyEventArgs e)
        {
            if (e.KeyCode.ToString() != "Return" && PenSizeDropBox.Focused == false &&
                (e.KeyCode.ToString() == "ShiftKey" || e.KeyCode.ToString() == "ControlKey"))
            {
                isKeyDown = false;
                if (SignaturePoints.Count > 0 && SignaturePoints[SignaturePoints.Count - 1] != null)
                {
                    SignaturePoints.Add(null);
                }
                e.Handled = true;

                ApplySmoothing_MouseClick(this, null);

                // Create an undo point on each keyup
                SetUndo(SignaturePoints);
            }
        }

        // Clear the drawing but not the Undo/Redo points
        private void Clear_MouseClick(object sender, MouseEventArgs e)
        {
            SignaturePixels = new System.Drawing.Bitmap(DrawingWidth, DrawingHeight);
            InitiatilizeSignaturePixels(ref SignaturePixels);

            SignaturePoints.Clear();
            SignaturePoints.Add(null);

            Signature.Image = SignaturePixels;

            SetUndo(SignaturePoints);
        }

        private void PenSizeDropBox_TextChanged(object sender, EventArgs e)
        {
            if (PenSizeDropBox.Text.Length > 0)
            {
                if (PenSizeDropBox.Text == "Hide")
                {
                    ShowPointLocation.Hide();
                    PenSizeDropBox.Text = UserSettings.Default.LineThickness.ToString();

                    if (mediaCtrl != null)
                    {
                        int state1, state2, state3;
                        state2 = 500;
                        state3 = 0;
                        state1 = mediaCtrl.GetState(state2, out state3);

                        if (state3 == 2)
                        {
                            mediaCtrl.Stop();
                        }
                    }

                    UserSettings.Default.ActivateWebcam = false;
                    UserSettings.Default.Save();
                }
                else if (PenSizeDropBox.Text == "Show")
                {
                    ShowPointLocation.Show();
                    PenSizeDropBox.Text = UserSettings.Default.LineThickness.ToString();
                    if (mediaCtrl != null)
                    {
                        int state1, state2, state3;
                        state2 = 500;
                        state3 = 0;
                        state1 = mediaCtrl.GetState(state2, out state3);

                        if (state3 != 2)
                        {
                            mediaCtrl.Run();
                        }
                    }
                    else
                    {
                        InitializeCamera();
                    }

                    UserSettings.Default.ActivateWebcam = true;
                    UserSettings.Default.Save();
                }
                else
                {
                    try
                    {
                        float SelectedThickness = float.Parse(PenSizeDropBox.Text);
                        if (SelectedThickness > 0)
                        {
                            UserSettings.Default.LineThickness = SelectedThickness;
                            UserSettings.Default.Save();
                            PenThickness = SelectedThickness;
                        }
                        InitiatilizeSignaturePixels(ref SignaturePixels);
                        RedrawPixels(ref SignaturePixels);
                    }
                    catch (Exception)
                    {
                        if (PenSizeDropBox.Text.Length >= 1)
                            PenSizeDropBox.Text = PenSizeDropBox.Text.Substring(0, PenSizeDropBox.Text.Length - 1);
                    }
                }
            }
        }

        private void PenSizeDropBox_DropDownClosed(object sender, EventArgs e)
        {
            Signature.Focus();
        }

        private void PenSizeDropBox_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Enter)
            {
                Signature.Focus();
                e.Handled = true;
            }
        }

        private void splitContainer1_Panel1_MouseClick(object sender, MouseEventArgs e)
        {
            splitContainer1.Panel1.Focus();
        }

        private void Signature_MouseClick(object sender, MouseEventArgs e)
        {
            Signature.Focus();
        }

        void PenColorButton_Click(object sender, System.EventArgs e)
        {
            ColorButton callingButton = (ColorButton)sender;
            Point p = new Point(groupBoxPen.Location.X + callingButton.Left, groupBoxPen.Location.Y + callingButton.Top + callingButton.Height);
            p = PointToScreen(p);

            ColorPaletteDialog clDlg = new ColorPaletteDialog(p.X, p.Y);

            clDlg.ShowDialog();

            if (clDlg.DialogResult == DialogResult.OK)
            {
                ColorConverter colconv = new ColorConverter();
                UserSettings.Default.LineColor = colconv.ConvertToString(clDlg.Color);
                UserSettings.Default.Save();
                SignatureColorSystem = (Color)colconv.ConvertFromString(UserSettings.Default.LineColor);
                SignatureColor.blue = (byte)clDlg.Color.B;
                SignatureColor.green = (byte)clDlg.Color.G;
                SignatureColor.red = (byte)clDlg.Color.R;

                InitiatilizeSignaturePixels(ref SignaturePixels);
                RedrawPixels(ref SignaturePixels);
            }

            callingButton.CenterColor = SignatureColorSystem;

            Invalidate();

            clDlg.Dispose();
        }

        void BackgroundColorButton_Click(object sender, System.EventArgs e)
        {
            ColorButton callingButton = (ColorButton)sender;
            Point p = new Point(groupBoxBackground.Location.X + callingButton.Left, groupBoxBackground.Location.Y + callingButton.Top + callingButton.Height);
            p = PointToScreen(p);

            ColorPaletteDialog clDlg = new ColorPaletteDialog(p.X, p.Y);

            clDlg.ShowDialog();

            if (clDlg.DialogResult == DialogResult.OK)
            {
                ColorConverter colconv = new ColorConverter();
                UserSettings.Default.BackColor = colconv.ConvertToString(clDlg.Color);
                UserSettings.Default.Save();
                backColorSystem = (Color)colconv.ConvertFromString(UserSettings.Default.BackColor);
                backColor.blue = (byte)clDlg.Color.B;
                backColor.green = (byte)clDlg.Color.G;
                backColor.red = (byte)clDlg.Color.R;

                InitiatilizeSignaturePixels(ref SignaturePixels);
                RedrawPixels(ref SignaturePixels);
            }

            callingButton.CenterColor = backColorSystem;

            Invalidate();

            clDlg.Dispose();
        }

        public void MainForm_Deactivate(object sender, EventArgs e)
        {
            //if (FineTuneTC != null && GetForegroundWindow() == FineTuneTC.Handle)
            //    FormActivated = true;
            //else
            //    FormActivated = false;
        }

        public void MainForm_Activated(object sender, EventArgs e)
        {
            FormActivated = true;
        }

        private void MakeBackTransparent_CheckedChanged(object sender, EventArgs e)
        {
            if (MakeBackTransparent.Checked == true)
            {
                UserSettings.Default.TransparentOnSave = true;
            }
            else
            {
                UserSettings.Default.TransparentOnSave = false;
            }
            UserSettings.Default.Save();
        }

        private void HelpAbout_MouseClick(object sender, MouseEventArgs e)
        {
            System.Diagnostics.Process.Start("http://webcamsignature.wikidot.com/help-faq");
        }

        #endregion

        #region For Webcam Signature 2.0
        // For Webcam Signature 2.0:
        // Explore the possiblity of tracking brightness
        // Brightness = (299 * red + 587 * green + 114 * blue) / 1000

        // Code the dialation and erosion functions to smooth out the pixel blobs

        // Fix the finicky tracking at the edges of the window.
        // One possiblity is to reduce the window size of the program.
        // Another possiblity is code a scaling function that maps a pixel from
        // the webcam image to a pixel on the signature area.

        long MouseMoveCounter = 0;
        private void Signature_MouseMove(object sender, MouseEventArgs e)
        {
            Point pt = e.Location;
            if (MouseIsDown && !(SignaturePoints.Count > 0 && SignaturePoints[SignaturePoints.Count - 1] != null && pt == (Point)SignaturePoints[SignaturePoints.Count - 1]))
            {
                long remainder = 0;
                Math.DivRem(MouseMoveCounter, 2, out remainder);
                if (remainder == 0)
                {
                    SignaturePoints.Add(pt);

                    // Do a rough representation of the signature while the user draws it
                    if (SignaturePoints.Count >= 2 && SignaturePoints[SignaturePoints.Count - 2] != null &&
                        SignaturePoints[SignaturePoints.Count - 1] != null)
                    {
                        DrawLine((Point)SignaturePoints[SignaturePoints.Count - 2], (Point)SignaturePoints[SignaturePoints.Count - 1],
                            ref SignaturePixels, true);
                    }
                    else if (SignaturePoints.Count >= 1 && SignaturePoints[SignaturePoints.Count - 1] != null)
                    {
                        DrawAPoint(pt, ref SignaturePixels, true);
                    }
                }
                else
                {
                    // Do a rough representation of the signature while the user draws it
                    //if (SignaturePoints.Count >= 1 && SignaturePoints[SignaturePoints.Count - 1] != null)
                    //{
                    //    DrawLine(pt, (Point)SignaturePoints[SignaturePoints.Count - 1],
                    //        ref SignaturePixels, true);
                    //}
                    //else if (SignaturePoints.Count >= 1 && SignaturePoints[SignaturePoints.Count - 1] != null)
                    //{
                    //    DrawAPoint(pt, ref SignaturePixels, true);
                    //}
                }



                MouseMoveCounter++;
            }

            //Point pt = new Point();
            //pt.X = e.X;
            //pt.Y = e.Y;
            ////Console.WriteLine(pt);

            //bool inZone = false;
            //Point inPt = new Point();

            //for (int i = pt.X - 7; i <= pt.X + 7 && i >= 0 && i < Signature.Size.Width; i++)
            //{
            //    for (int j = pt.Y - 7; j <= pt.Y + 7 && j >= 0 && j < Signature.Size.Height; j++)
            //    {
            //        inPt.X = i;
            //        inPt.Y = j;
            //        //Console.WriteLine("comparing");
            //        if (SignaturePoints.Contains(inPt))
            //        {
            //            //Console.WriteLine("in zone");
            //            inZone = true;
            //            break;
            //        }
            //    }
            //    if (inZone)
            //        break;
            //}

            //if (inZone)
            //{
            //    Signature.Cursor = Cursors.Hand;
            //}
            //else
            //{
            //    Signature.Cursor = Cursors.Default;
            //}
        }

        bool MouseIsDown = false;
        private void Signature_MouseUp(object sender, MouseEventArgs e)
        {
            MouseIsDown = false;

            if (SignaturePoints.Count > 0 && SignaturePoints[SignaturePoints.Count - 1] != null)
            {
                SignaturePoints.Add(null);
            }

            ApplySmoothing_MouseClick(this, null);

            // Create an undo point on each keyup
            SetUndo(SignaturePoints);

            MouseMoveCounter = 0;
        }

        private void Signature_MouseDown(object sender, MouseEventArgs e)
        {
            MouseIsDown = true;          
        }

        private void Signature_MouseLeave(object sender, EventArgs e)
        {
        }

        #endregion

        private void MainForm_SizeChanged(object sender, EventArgs e)
        {
            splitContainer1.SplitterDistance = 150;
        }

        private void WebcamMouseCB_CheckedChanged(object sender, EventArgs e)
        {
            if (WebcamMouseCB.Checked == true)
            {
                ShowPointLocation.Location = new Point(-100, -100);
                WebcamMouseMode = true;
            }
            else
            {
                WebcamMouseMode = false;
            }
        }
    }
}
